import 'package:contentsquare/src/modules/screengraph/render_object_extension.dart';
import 'package:flutter/widgets.dart';

/// An enum describing a boolean with a third value: unknown
enum MaybeResult {
  /// true
  yes,

  /// false
  no,

  /// unknown
  unknown,
}

/// An extension on [Element] that provides some utility methods
extension ElementExtension on Element {
  /// Returns the [RenderObject] of the given [Element] if any
  ///
  /// The difference between this and [Element.renderObject] is that
  /// this does not walk the tree to get the 1st [RenderObject] but only
  /// returns the [RenderObject] attached directly to the current [Element]
  /// which only happens if it's a [RenderObjectElement]
  RenderObject? get directRenderObject {
    final _this = this;
    if (_this is RenderObjectElement) {
      return _this.renderObject; // Not null !
    }

    return null;
  }

  /// Whether this element contains the [globalPosition]
  ///
  /// Returns [MaybeResult.unknown] if we can't say if it contains the
  /// [globalPosition] (because this [Element] is not associated with a
  /// [RenderObject])
  MaybeResult containsGlobalPosition(Offset globalPosition) {
    final _directRenderObject = directRenderObject;

    // If this element has no direct render object, we can't say if it contains
    // the globalPosition
    if (_directRenderObject == null) {
      return MaybeResult.unknown;
    }

    final globalFrame = _directRenderObject.globalFrame;

    if (globalFrame != null && globalFrame.contains(globalPosition)) {
      return MaybeResult.yes;
    }

    return MaybeResult.no;
  }

  /// Returns the name of the [Widget] associated with this element
  String get widgetName => widget.runtimeType.toString();
}
