import 'dart:collection';

import 'package:contentsquare/src/modules/screengraph/widgets_binding.dart';
import 'package:flutter/widgets.dart';

import 'element_extension.dart';
import 'element_visitor.dart';

/// Returns the root [Element] of the tree.
Element getRootElement() {
  return WidgetsBindingExtension.safeInstance!.renderViewElement!;
}

/// Return the path to set to the root [Element] of the tree.
String getRootElementPath() {
  return '${getRootElement().widgetName}:eq(0)';
}

/// Returns the [ElementNode] tree calculated from the root of the app
ElementNode getRootElementTree() {
  return _getElementNode(elementVisitor: CSElementVisitor.root());
}

/// Represents a node in the tree.
///
/// The tree is a mirror of the [Element] tree, where additional information
/// have been gathered while useless info has been removed.
class ElementNode {
  /// Creates an [ElementNode]
  ElementNode({
    required this.widget,
    required this.renderObject,
    required this.path,
    required this.children,
  });

  /// The [Widget] of the [Element] that this node represents
  final Widget widget;

  /// The [RenderObject] of the [Element] that this node represents if any
  ///
  /// The [RenderObject] is an [ElementExtension.directRenderObject]
  final RenderObject? renderObject;

  /// The path to this [RenderObject] from the root of the tree
  final String path;

  /// The children of this node, based on the children of the associated [Element]
  final UnmodifiableListView<ElementNode> children;

  @override
  String toString() {
    return 'ElementNode{widget: ${widget.runtimeType}, renderObject: ${renderObject.runtimeType}, path: $path, children: $children}';
  }
}

/// Returns the visible [ElementNode]s tree with [elementVisitor.element] as the
/// root
ElementNode _getElementNode({required CSElementVisitor elementVisitor}) {
  final List<ElementNode> childrenNodes = <ElementNode>[];
  elementVisitor.visitChildren(
    visit: (childElementVisitor) {
      childrenNodes.add(_getElementNode(elementVisitor: childElementVisitor));

      // Never stop visiting children
      return false;
    },
  );

  return ElementNode(
    widget: elementVisitor.element.widget,
    renderObject: elementVisitor.element.directRenderObject,
    path: elementVisitor.tvp,
    children: UnmodifiableListView(childrenNodes),
  );
}
