import 'package:flutter/rendering.dart';
import 'package:flutter/widgets.dart';

import 'element_extension.dart';
import 'element_tree.dart';

/// A class containing an element and which allows the visitation of its
/// children
///
///
/// Troubleshooting:
/// [CSElementVisitor] and [ElementVisitor] are DIFFERENT:
///   - [CSElementVisitor] is a class containing an element and which allows the visitation of its children
///   - [ElementVisitor] is a function typedef that is used to describe the callback to give to [Element.visitChildren]
class CSElementVisitor {
  /// The element associated with this visitor
  final Element element;

  /// The Target View Path of this element
  ///
  /// See [_createTVPFromParent] to understand how a TVP is build from the TVP
  /// of its parent and sibling elements
  final String tvp;

  /// Describes whether the children of this element might be _OverlayEntryWidget
  /// created by the navigator.
  ///
  /// This is used to determine the children TVPs.
  ///
  ///
  /// 1. At the root of the tree, children cannot be created by a navigator so
  ///    this is false until a child is a [Navigator]
  /// 2. From the [Navigator], this stays true until a child is an [_OverlayEntryWidget]
  /// 3. From an [_OverlayEntryWidget], this stays false until a child is a [Navigator]
  /// 4. Repeat from 2.
  ///
  /// Here is an example of the tree:
  ///                                     Widgets:  ...  -> Navigator -> Overlay -> _Theatre -> _OverlayEntryWidget ->  ...
  /// _childrenMayBeOverlayEntryWidgetOfNavigator: false ->   true    ->  true   ->   true   ->       false         -> false
  final bool _childrenMayBeOverlayEntryWidgetOfNavigator;

  /// Calls the [visit] function on each child of this element until the [visit]
  /// function returns true.
  void visitChildren({
    required bool Function(CSElementVisitor childElementVisitor) visit,
  }) {
    // [RenderSliverMultiBoxAdaptor] have a special treatment because they
    // have several children which can be recycled.
    //
    // Instead of using an index based on the type, we use the index value of
    // each child directly (using the fact that they are [RenderIndexedSemantics],
    // which holds an index). This ensures that the the TVP is consistent across
    // different scroll positions.
    final isRenderSliverMultiBoxAdaptor =
        element.renderObject is RenderSliverMultiBoxAdaptor;

    // The [typeCounts] map contains the number of times each [RenderObject] type
    // appears in the childrenElement, which is used to compute the path of each
    // [RenderObjectNode]
    final Map<Type, int> typeCounts = <Type, int>{};
    var _shouldVisit = true;
    element.visitChildren((childElement) {
      if (_shouldVisit) {
        final widgetType = childElement.widget.runtimeType;

        // Update the count based on the widget type
        final count = (typeCounts[widgetType] ?? 0) + 1;
        typeCounts[widgetType] = count;

        final int index;
        if (isRenderSliverMultiBoxAdaptor) {
          // Use the index of the [RenderIndexedSemantics] so that the TVP is
          // consistent across different scroll positions
          final childRenderObject = childElement.renderObject;
          if (childRenderObject is RenderIndexedSemantics) {
            index = childRenderObject.index;
          } else {
            index = count - 1;
          }
        } else {
          index = count - 1;
        }

        // We can't test the type of [_OverlayEntryWidget] because it's private
        final childIsOverlayEntryWidget =
            childElement.widget.runtimeType.toString() == '_OverlayEntryWidget';

        final childIsOverlayEntryWidgetOfNavigator =
            _childrenMayBeOverlayEntryWidgetOfNavigator &&
                childIsOverlayEntryWidget;

        final childPath = _createTVPFromParent(
          parentTVP: tvp,
          childElement: childElement,
          childIndex: index,
          shouldIndexTVP: !childIsOverlayEntryWidgetOfNavigator,
        );

        final childChildrenMayBeOverlayEntryWidgetOfNavigator = () {
          if (childIsOverlayEntryWidgetOfNavigator) {
            return false;
          }

          return _childrenMayBeOverlayEntryWidgetOfNavigator ||
              childElement.widget is Navigator;
        }();

        _shouldVisit = !visit(
          CSElementVisitor._(
            element: childElement,
            tvp: childPath,
            childrenMayBeOverlayEntryWidgetOfNavigator:
                childChildrenMayBeOverlayEntryWidgetOfNavigator,
          ),
        );
      }
    });
  }

  CSElementVisitor._({
    required this.element,
    required this.tvp,
    required bool childrenMayBeOverlayEntryWidgetOfNavigator,
  }) : _childrenMayBeOverlayEntryWidgetOfNavigator =
            childrenMayBeOverlayEntryWidgetOfNavigator;

  /// Creates the [CSElementVisitor] associated with the root element of the tree
  CSElementVisitor.root()
      : element = getRootElement(),
        tvp = getRootElementPath(),
        _childrenMayBeOverlayEntryWidgetOfNavigator = false;

  /// Creates the [CSElementVisitor] with the given element
  ///
  /// This should only be used for testing, otherwise use [ElementVisitor.root]
  /// to create the root element visitor and [visitChildren] to build its
  /// children
  ///
  /// We don't want the full constructor to be used directly so we explicitly
  /// name it createForTesting to make it clear that it should not be used
  /// directly in production code
  @visibleForTesting
  CSElementVisitor.createForTesting({
    required this.element,
    required this.tvp,
    required bool childrenMayBeOverlayEntryWidgetOfNavigator,
  }) : _childrenMayBeOverlayEntryWidgetOfNavigator =
            childrenMayBeOverlayEntryWidgetOfNavigator;

  /// Tests if two [CSElementVisitor]s are equal
  @visibleForTesting
  bool equals(CSElementVisitor other) =>
      identical(this, other) ||
      runtimeType == other.runtimeType &&
          element == other.element &&
          tvp == other.tvp &&
          _childrenMayBeOverlayEntryWidgetOfNavigator ==
              other._childrenMayBeOverlayEntryWidgetOfNavigator;
}

/// The TVP of [childElement] is the concatenation of:
///   1. The TVP of the parent
///   2. The runtimeType of the widget of the [childElement]
///   3. The index of [childElement] in the parent's children list of the same type
///
/// Note: [childIndex] is ignored if [shouldIndexTVP] is true.
/// Using this allows a screen pushed at different places in the [Navigator]
/// stack to always have the same children TVPs
/// It is ok not to use it because the OverlayEntryWidget of navigators cover up
/// the OverlayEntryWidget below which will therefore never be a target.
String _createTVPFromParent({
  required String parentTVP,
  required Element childElement,
  required int childIndex,
  required bool shouldIndexTVP,
}) {
  if (!shouldIndexTVP) {
    return '$parentTVP>${childElement.widgetName}';
  }

  return '$parentTVP>${childElement.widgetName}:eq($childIndex)';
}
