import 'dart:collection';
import 'dart:developer';

import 'package:flutter/rendering.dart';
import 'package:flutter/widgets.dart';

import '../../modules/screengraph/widgets_binding.dart';
import 'compression.dart';
import 'element_extension.dart';
import 'element_visitor.dart';
import 'painted_render_object_tree.dart';
import 'render_object_extension.dart';
import 'unmodifiable_list_view_extension.dart';

/// The TVP from the root of the tree to the [RenderObject] at [globalHitPosition].
///
///
/// This algorithm does the following:
///   1. Find the deepest [RenderObject] that respond to a hit at [globalHitPosition].
///   2. Find the [Element] that corresponds to the [RenderObject] (if none,
///      return empty).
///   3. Find the leaves of the screengraph that are below the found [Element]
///      and which frame includes the [globalHitPosition].
///
///
/// We might have no match if at some point in the hit test, a renderObject
/// populated the [BoxHitTestResult] while the [globalHitPosition] was outside of
/// its globalFrame.
/// To date, we have not found any widget capable of doing this.
///
/// We might have several matches because: the element that responded to the hit
/// test at [globalHitPosition] might have several children that:
///   1. Have a frame which includes [globalHitPosition]
///   2. Are part of the sent screengraph.
/// For example:
/// ┌──────────────────────────────────┐
/// │         Gesture detector         │
/// │                                  │
/// │  ┌────────────────────────────┐  │
/// │  │  TextWidget1               │  │
/// │  │                            │  │
/// │  │             ┌────────────┐ │  │
/// │  │             │     x ◄────┼─┼──┼─── Hit position
/// │  └─────────────┼────────────┼─┘  │
/// │                │            │    │
/// │                │ TextWidget2│    │
/// │                └────────────┘    │
/// └──────────────────────────────────┘
/// Here Gesture detector is the deepest widget that responded to the hit test.
/// TextWidget1 and TextWidget2 are children of Gesture detector that:
///   1. have a frame which include the hit position
///   2. are part of the screengraph (because their frame is different from the
///      one of the Gesture detector).
/// Therefore, 2 TVPs will be returned (one to TextWidget1 and one to TextWidget2).
UnmodifiableListView<String> getTVPAtHitPosition(Offset globalHitPosition) {
  final hitRenderObject = _getHitRenderObject(globalHitPosition);

  final hitRenderObjectElementParams = _pathToHitRenderObject(
    renderObject: hitRenderObject,
    params: _HitElementParams.root(globalHitPosition: globalHitPosition),
  );

  // If [hitRenderObject] could not be found
  if (hitRenderObjectElementParams == null) {
    log(
      'Could not find the element associated with the RenderObject $hitRenderObject that was found for the gesture at $globalHitPosition',
    );
    return UnmodifiableListView([]);
  }

  final hitLeavesRenderObjectsPaths = _leavesRenderObjectsPathsAt(
    params: hitRenderObjectElementParams,
  );

  // See [_leavesRenderObjectsPathsAt] to understand why we need to remove
  // duplicates and inclusions.
  final uniqueHitLeavesRenderObjectsPaths =
      hitLeavesRenderObjectsPaths.removeDuplicatesAndInclusions();

  final uniqueCompressedHitLeavesRenderObjectsPaths =
      uniqueHitLeavesRenderObjectsPaths.map(compressTVP).toList();

  return UnmodifiableListView(uniqueCompressedHitLeavesRenderObjectsPaths);
}

/// Returns the deepest render object which:
///   - Contains the given point ([position])
///   - Reacts to hitTest
///
/// The function takes into account what is currently on the screen by leveraging
/// the hitTest method used by the flutter framework to find which renderObject
/// has been hit.
/// Therefore the returned render object should be consistent with what is visible
/// on the screen.
///
/// WARNING: This function will only return the deepest render object which
/// reacts to the hitTest. That means that the returned render object is not
/// actually the deepest render object containing the point at [position].
RenderObject _getHitRenderObject(Offset position) {
  final rootRenderObject = WidgetsBindingExtension.safeInstance!.renderView;

  // result will be populated with the list of render objects that the hit
  // test intersects.
  final result = BoxHitTestResult();
  rootRenderObject.hitTest(result, position: position);

  // We don't need to check for null because [RenderView] will always match.
  return result.path
      .firstWhere((element) => element.target is RenderObject)
      .target as RenderObject;
}

/// Returns the path leading to the [renderObject]
///
///
/// [params] contains information about:
///   - the element from which we should start the search for the hit render object.
///   - the position of the hit
///
///
/// The returned object is a [_HitElementParams] that contains the path to the
/// [renderObject] that was hit from the root of the tree.
_HitElementParams? _pathToHitRenderObject({
  required RenderObject renderObject,
  required _HitElementParams params,
}) {
  // If the offset is NOT inside the globalFrame of the element, we consider
  // that neither this element nor its children is the hit [renderObject].
  //
  //
  // We use this for performance reasons: avoid visiting parts of the tree
  // that are not around the [globalHitPosition] allows to cut down large parts
  // of the tree, resulting in:
  //   - Faster algorithm
  //   - Lighter screengraph
  final containsGlobalPosition = params.elementVisitor.element
      .containsGlobalPosition(params.globalHitPosition);
  if (containsGlobalPosition == MaybeResult.no) {
    return null;
  }

  // Use [visitElementChildren] so that the path is iso with the one calculated
  // in [getRootElementTree].
  _HitElementParams? childMatch;
  params.elementVisitor.visitChildren(
    visit: (childElementVisitor) {
      childMatch = _pathToHitRenderObject(
        renderObject: renderObject,
        params: params.buildChild(childElementVisitor: childElementVisitor),
      );

      // Stop if we found a match.
      return childMatch != null;
    },
  );

  // If we found a match in the children, return it.
  if (childMatch != null) {
    return childMatch;
  }

  // If the element holds the render object, return it
  if (params.elementVisitor.element.directRenderObject == renderObject) {
    return params;
  }

  // If neither the children nor this element hold the render object, return null.
  return null;
}

/// Searches for the TVP of the leaf elements at the given
/// [_HitElementParams.globalHitPosition].
/// An element is a leaf if it has no children at the given [_HitElementParams.globalHitPosition].
///
///
/// This function might return duplicates if two children use the
/// [_HitElementParams.nearestScreengraphElementTVP] of the parent.
///
/// This function might return element included in one another if one child uses
/// the same [_HitElementParams.nearestScreengraphElementTVP] as the parent while the other does not.
UnmodifiableListView<String> _leavesRenderObjectsPathsAt({
  required _HitElementParams params,
}) {
  // If the globalPosition is NOT inside the globalFrame of the element, neither
  // the element nor its children are the ones that we are looking for.
  final containsGlobalPosition = params.elementVisitor.element
      .containsGlobalPosition(params.globalHitPosition);
  if (containsGlobalPosition == MaybeResult.no) {
    return UnmodifiableListView([]);
  }

  // Use [visitElementChildren] so that the path is iso with the one calculated
  // in [getRootElementTree].
  final childrenMatches = <String>[];
  params.elementVisitor.visitChildren(
    visit: (childElementVisitor) {
      // Recurse on the child element.
      childrenMatches.addAll(
        _leavesRenderObjectsPathsAt(
          params: params.buildChild(childElementVisitor: childElementVisitor),
        ),
      );

      // Search all children.
      return false;
    },
  );

  // If we found a match in the children, return it.
  if (childrenMatches.isNotEmpty) {
    return UnmodifiableListView(childrenMatches);
  }

  // If no children matched and this element contains the globalPosition,
  // return the path of this element.
  //
  // We need to check the value of [containsGlobalPosition] because it might be
  // [MaybeEnum.unknown].
  if (containsGlobalPosition == MaybeResult.yes) {
    return UnmodifiableListView([params.nearestScreengraphElementTVP]);
  }

  return UnmodifiableListView([]);
}

/// A class which can be used as a param to search for elements that was hit
/// at [globalHitPosition], starting from the element [elementVisitor]
class _HitElementParams {
  /// Returns
  factory _HitElementParams.root({required Offset globalHitPosition}) {
    // [globalFrame] is not null because we assume that the root render object
    // has a size. This is the case as soon as runApp is called.
    final rootGlobalFrame = getRootRenderObject().globalFrame!;

    final rootElementVisitor = CSElementVisitor.root();

    return _HitElementParams._(
      globalHitPosition: globalHitPosition,
      elementVisitor: rootElementVisitor,
      elementGlobalFrame: rootGlobalFrame,
      nearestScreengraphElementTVP: rootElementVisitor.tvp,
    );
  }

  /// [childElementVisitor] must be a child of [elementVisitor]
  _HitElementParams buildChild({required CSElementVisitor childElementVisitor}) {
    final childGlobalFrame =
        childElementVisitor.element.directRenderObject?.globalFrame;

    final isChildIncludedInScreengraph = getIsChildIncludedInScreengraph(
      childFrame: childGlobalFrame,
      parentFrame: elementGlobalFrame,
    );

    return _HitElementParams._(
      globalHitPosition: globalHitPosition,
      elementVisitor: childElementVisitor,
      // Elements included in the screengraph have a non-null globalFrame.
      elementGlobalFrame:
          isChildIncludedInScreengraph ? childGlobalFrame! : elementGlobalFrame,
      nearestScreengraphElementTVP: isChildIncludedInScreengraph
          ? childElementVisitor.tvp
          : nearestScreengraphElementTVP,
    );
  }

  /// The position (in the global coordinate system) at which the hit was
  /// performed
  final Offset globalHitPosition;

  /// The element from which to start the search for a hit element
  ///
  /// Note: we do not guarantee that this element globalFrame contains
  /// [globalHitPosition]
  final CSElementVisitor elementVisitor;

  /// The global frame of the [RenderObject] associated with:
  ///   - [elementVisitor] if it has a [RenderObject]
  ///   - the nearest parent of [elementVisitor] which has a [RenderObject] otherwise
  final Rect elementGlobalFrame;

  /// The path of either:
  ///   - [elementVisitor] if it was sent in the screengraph
  ///   - or the nearest parent of [elementVisitor] which was sent in the screengraph
  final String nearestScreengraphElementTVP;

  /// Don't create instances of this class directly.
  ///
  /// Instead, use [_HitElementParams.root] or [_HitElementParams.buildChild]
  _HitElementParams._({
    required this.globalHitPosition,
    required this.elementVisitor,
    required this.elementGlobalFrame,
    required this.nearestScreengraphElementTVP,
  });
}
