import 'dart:collection';
import 'dart:convert';

import 'package:flutter/rendering.dart';
import 'package:flutter/widgets.dart';

import 'compression.dart';
import 'element_tree.dart';
import 'render_object_extension.dart';
import 'widgets_binding.dart';

/// Returns the root [RenderObject] of the [RenderObject] tree.
RenderView getRootRenderObject() {
  return WidgetsBindingExtension.safeInstance!.renderView;
}

/// The screengraph is formed by:
///   1. Getting the element tree in the [ElementNode] format: [getRootElementTree]
///   2. Converting the [ElementNode]s to [PaintedRenderObjectNode]s by walking
///      the painting tree: [_getRootPaintedRenderObjectTree]
///   3. Filtering the [PaintedRenderObjectNode]s into a screengraph by only
///      keeping the [PaintedRenderObjectNode]s that have a frame different from
///      their parent: [PaintedRenderObjectNode.toScreengraph]
/// Here is a diagram of what is happening:
/// ┌──────────────────────────────────────┐
/// │ Element                              │
/// │ tree      ┌──────────────┐           │
/// │           │ Element Node │           │
/// │           │              │           │
/// │           │         20,20│           │
/// │           └──────┬───────┘           │
/// │                  │                   │
/// │           ┌──────┴───────┐           │
/// │           │ Element Node │           │
/// │           │              │           │
/// │           │ RO1     20,20│           │
/// │           └──────┬───────┘           │
/// │                  │                   │
/// │           ┌──────┴───────┐           │
/// │           │ Element Node │           │
/// │           │              │           │
/// │           │ RO2     20,20│           │
/// │           └──────┬───────┘           │
/// │                  │                   │
/// │        ┌─────────┴─────────┐         │
/// │        │                   │         │
/// │ ┌──────┴───────┐    ┌──────┴───────┐ │
/// │ │ Element Node │    │ Element Node │ │
/// │ │              │    │              │ │
/// │ │ RO3     15,15│    │ RO4     05,05│ │
/// │ └──────────────┘    └──────────────┘ │
/// │                                      │
/// └────┬────────────────────────────┬────┘
///      │ Remove the elements which  │
///      │  *don't have a RO          │
///      ▼  *RO have not been painted ▼
/// ┌──────────────────────────────────────┐
/// │ Painted                              │
/// │ render        ┌─────────────────┐    │
/// │ object        │ Painted RO Node │    │
/// │ tree          │                 │    │
/// │               │ RO1        20,20│    │
/// │               └────────┬────────┘    │
/// │                        │             │
/// │               ┌────────┴────────┐    │
/// │               │ Painted RO Node │    │
/// │               │                 │    │
/// │               │ RO2        20,20│    │
/// │               └────────┬────────┘    │
/// │                        │             │
/// │              ┌─────────┘             │
/// │              │                       │
/// │      ┌───────┴─────────┐             │
/// │      │ Painted RO Node │             │
/// │      │                 │             │
/// │      │ RO3        15,15│             │
/// │      └─────────────────┘             │
/// │                                      │
/// └───┬─────────────────────────────┬────┘
///     │  Remove RO which frame is   │
///     ▼   the same as its parent    ▼
/// ┌───────────────────────────────────────┐
/// │ Screengraph                           │
/// │            ┌────────────────────────┐ │
/// │            │ Painted RO Node (json) │ │
/// │            │                        │ │
/// │            │ RO1               20,20│ │
/// │            └───────┬────────────────┘ │
/// │                    │                  │
/// │                    │                  │
/// │                    │                  │
/// │                ┌───┘                  │
/// │                │                      │
/// │        ┌───────┴────────────────┐     │
/// │        │ Painted RO Node (json) │     │
/// │        │                        │     │
/// │        │ RO3               15,15│     │
/// │        └────────────────────────┘     │
/// │                                       │
/// └───────────────────────────────────────┘
///
/// This gives us a screengraph that has all the elements painted on the screen
/// which have a frame different from their parent.
Future<String> getScreengraphFromRoot({
  required String nativePath,
  required double pixelRatio,
}) async {
  final rootPaintedRenderObjectTree = await _getRootPaintedRenderObjectTree(
  );

  final screengraph = rootPaintedRenderObjectTree.toScreengraph(
    pixelRatio: pixelRatio,
    // [parentGlobalFrame] is null because the root doesn't have a parent.
    parentGlobalFrame: null,
    nativePath: nativePath,
  );

  // Add the placeholder so that the native static snapshot can be inserted
  // screengraph[0] is the node corresponding to the root widget
  screengraph[0]['style']['bmp'] =
      'com_contentsquare_rootViewBitmapPlaceholder';

  return json.encode(screengraph);
}

/// Returns the [PaintedRenderObjectNode] tree calculated from the root of the
/// app
///
/// This is basically a tree of the [RenderObject]s that have been painted on the
/// screen where the root of the tree is the root [RenderObject] of the app.
///
Future<PaintedRenderObjectNode> _getRootPaintedRenderObjectTree() async {
  final rootRenderObject = getRootRenderObject();

  final rootElementTree = getRootElementTree();

  final rootPaintedRenderObjectNode = await _getPaintedRenderObjectNode(
    paintedRenderObject: rootRenderObject,
    elementTree: rootElementTree,
  );

  // We are guaranteed that this isn't null because the root element will always
  // have a local and global frame and therefore will always create a node.
  return rootPaintedRenderObjectNode!;
}

/// Node of a tree representing the [RenderObject] that are painted
class PaintedRenderObjectNode {
  /// Creates a [PaintedRenderObjectNode]
  PaintedRenderObjectNode({
    required this.renderObject,
    required this.globalFrame,
    required this.path,
    required this.widget,
    required this.children,
  });

  /// The [RenderObject] associated with this node
  final RenderObject renderObject;

  /// The size and position of [renderObject] w.r.t the screen
  final Rect globalFrame;

  /// The path to this [RenderObject] from the root of the tree
  final String path;

  /// The widget associated with this node
  final Widget widget;

  /// The children of this node
  ///
  /// This contains all the children of the [renderObject] that are being painted
  final UnmodifiableListView<PaintedRenderObjectNode> children;

  /// A name to easily identify this [ElementNode]
  String get name => '${widget.runtimeType}';

  /// Returns the json representation of this node and its children in the
  /// screengraph format
  ///
  /// Are only included the nodes that have a different frame than their parent
  ///
  /// [parentGlobalFrame] is the global frame of the parent node. This must be
  /// null if and only if this is the root node.
  ///
  /// [nativePath] must be the path to the Flutter view as viewed by the native SDK.
  UnmodifiableListView<Map<String, dynamic>> toScreengraph({
    required double pixelRatio,
    required Rect? parentGlobalFrame,
    required String nativePath,
  }) {
    assert(
      () {
        final isRoot = renderObject == getRootRenderObject();

        return (isRoot && parentGlobalFrame == null) ||
            (!isRoot && parentGlobalFrame != null);
      }(),
      'The parent global frame must be null if and only if this is the root node',
    );

    // [getIsChildIncludedInScreengraph] also tests if [globalFrame] is null
    // but this will never happen here because of how [PaintedRenderObjectNode]
    // are created
    final isChildIncludedInScreengraph = getIsChildIncludedInScreengraph(
      childFrame: globalFrame,
      parentFrame: parentGlobalFrame,
    );
    if (!isChildIncludedInScreengraph) {
      return UnmodifiableListView([
        for (final child in children)
          ...child.toScreengraph(
            pixelRatio: pixelRatio,
            parentGlobalFrame: globalFrame,
            nativePath: nativePath,
          )
      ]);
    }

    return UnmodifiableListView([
      {
        'id': name,
        'metadata': {
          'class_name': name,
          // Don't compress the native path because we don't have the native path
          // during gesture events
          'fullpath': nativePath + compressTVP(path),
        },

        /// "recordingId" should only be mandatory for session replay
        /// On the iOS SDK, the structure is used for both: snapshoting and session replay features
        /// so it ends up mandatory for the screengraph passed down to the ExternalBridgeInterface.takeSnapshot completion
        /// For this reason, we set temporarily a 'recordingId' value based on the widget's runtimeType.
        /// TODO: This will need to be checked and discussed when implementing session replay related features
        'recordingId': name,
        'style': {
          'height': globalFrame.height * pixelRatio,
          'width': globalFrame.width * pixelRatio,
          'x': globalFrame.left * pixelRatio,
          'y': globalFrame.top * pixelRatio,
          'visibility': true,
        },
        // TODO(lulupointu): implement when we handle webviews
        'format': 1,
        'children': [
          for (final child in children)
            ...child.toScreengraph(
              pixelRatio: pixelRatio,
              parentGlobalFrame: globalFrame,
              nativePath: nativePath,
            ),
        ],
      },
    ]);
  }
}

/// If the child frame is null or the same as its parent, return don't include the
/// child in the screengraph
///
/// This is used to:
///   - Reduce the number of nodes in the tree
///   - Create only interesting nodes (since our goal is zoning)
bool getIsChildIncludedInScreengraph({
  required Rect? parentFrame,
  required Rect? childFrame,
}) {
  return childFrame != null && parentFrame != childFrame;
}

/// Returns the [PaintedRenderObjectNode]s associated with [paintedRenderObject] and
/// its children
///
/// This is basically a tree of the [RenderObject]s that have been painted on the
/// screen where the root of the tree is [paintedRenderObject].
///
///
/// [elementTree] is the tree of [ElementNode]s of which the root element is the
/// element associated with [paintedRenderObject]. It is necessary to get information
/// about the [PaintedRenderObjectNode] and its children such as the path to
/// [paintedRenderObject].
///
///
/// Returns null if the frame of [paintedRenderObject] is null
Future<PaintedRenderObjectNode?> _getPaintedRenderObjectNode({
  required RenderObject paintedRenderObject,
  required ElementNode elementTree,
}) async {
  assert(
    elementTree.renderObject == paintedRenderObject,
    '	_getPaintedRenderObjectNode was called with a [ElementNode] in which the '
    'renderObject does not match the given [paintedRenderObject]',
  );

  final localFrame = paintedRenderObject.localFrame;
  final globalFrame = paintedRenderObject.globalFrame;

  // If localFrame is null, we assume that the [renderObject] is not painted
  // (since it has not been in the layout phase yet)
  //
  // These frame are null and non-null and the same time, we check both to
  // have compile time safety
  if (localFrame == null || globalFrame == null) {
    return null;
  }

  final childrenPaintedRenderObjectNodes = <PaintedRenderObjectNode>[];
  // We use [LayerHandle] to handle the [Layer]s as requested by the [Layer]
  // documentation.
  final containerLayerHandle = LayerHandle<ContainerLayer>();
  containerLayerHandle.layer = ContainerLayer();

  final paintingContext = _IsolatingPaintingContext(
    containerLayerHandle.layer!,
    localFrame,
  );

  // Call [renderObject.paint] so that the [renderObject] calls
  // [paintingContext.paintChild] for each of its children, allowing us to
  // get the childrenPaintedRenderObjects
  paintedRenderObject.paint(paintingContext, Offset.zero);

  // Dispose the layers, as requested by the [LayerHandle] documentation.
  containerLayerHandle.layer = null;

  // Get the children painted render objects nodes
  final childrenPaintedRenderObjects = paintingContext.paintedChildren;
  for (final childPaintedRenderObject in childrenPaintedRenderObjects) {
    final childElementTree = _getSubElementTree(
      renderObject: childPaintedRenderObject,
      elementTree: elementTree,
    );

    // It is possible that a [childPaintedRenderObject] is not present in the element tree
    //
    // For example a [RenderEditable] creates child render object that are not
    // associated with any element. Therefore they will be present in the
    // painted tree but not in the element tree.
    //
    // In this case we ignore the child since we can't associate it with any
    // path.
    if (childElementTree != null) {
      final childPaintedRenderObjectNode = await _getPaintedRenderObjectNode(
        paintedRenderObject: childPaintedRenderObject,
        elementTree: childElementTree,
      );

      if (childPaintedRenderObjectNode != null) {
        childrenPaintedRenderObjectNodes.add(childPaintedRenderObjectNode);
      }
    }
  }

  // The [renderObject] might have associated itself with our layer when we
  // called [renderObject.paint]
  //
  // Use [renderObject.markNeedsPaint] so that it re-associates itself with the
  // right layer during the next paint frame.
  paintedRenderObject.markNeedsPaint();

  return PaintedRenderObjectNode(
    renderObject: paintedRenderObject,
    globalFrame: globalFrame,
    widget: elementTree.widget,
    path: elementTree.path,
    children: UnmodifiableListView(childrenPaintedRenderObjectNodes),
  );
}

/// A [PaintingContext] which prevents the render object to render its children.
///
/// It also keeps track of the children that tried to be painted in [paintedChildren]
class _IsolatingPaintingContext extends PaintingContext {
  _IsolatingPaintingContext(
    ContainerLayer containerLayer,
    Rect estimatedBounds, {
    this.onPaintChild,
  }) : super(containerLayer, estimatedBounds);

  /// The children that the render object tried to paint in [paintChild]
  UnmodifiableListView<RenderObject> get paintedChildren =>
      UnmodifiableListView<RenderObject>(_paintedChildren);

  /// The list of [RenderObject] that where given during calls to [paintChild].
  ///
  /// This is populated each time [paintChild] is called.
  ///
  /// WARNING: This is never reset, which means that if you call [paintChild]
  /// on multiple [RenderObject]s, all their painted children will be contained
  /// in this list.
  final List<RenderObject> _paintedChildren = <RenderObject>[];

  /// This is what is called by a [renderObject] to paint a child.
  @override
  void paintChild(RenderObject child, Offset offset) {
    _paintedChildren.add(child);
    onPaintChild?.call(child);

    // Don't call super to avoid painting the child
  }

  /// A callback used in child context to report a child that tried to be painted.
  ///
  /// Read [createChildContext] for more information.
  final void Function(RenderObject child)? onPaintChild;

  /// Some render object use [_IsolatingPaintingContext.pushLayer] to paint their
  /// children. This calls [createChildContext] to create a new context on which
  /// the render object will paint its child.
  /// Therefore we need to also return an [_IsolatingPaintingContext] which will
  /// report to this [_IsolatingPaintingContext] the children that tried to be
  /// painted.
  @override
  _IsolatingPaintingContext createChildContext(
    ContainerLayer childLayer,
    Rect bounds,
  ) {
    return _IsolatingPaintingContext(
      childLayer,
      bounds,
      onPaintChild: _paintedChildren.add,
    );
  }
}

/// Return the sub-tree of the element tree of where the parent element
/// is associated with [renderObject]
///
/// If no element where found to be associated with [renderObject], return null.
ElementNode? _getSubElementTree({
  required RenderObject renderObject,
  required ElementNode elementTree,
}) {
  final rootNode = elementTree;

  // Do a breadth-first search to find [renderObject] in the tree which root
  // element is [rootNode].
  final nodes = List.of([rootNode]);
  while (nodes.isNotEmpty) {
    final node = nodes.removeAt(0);
    if (node.renderObject == renderObject) {
      return node;
    }

    nodes.addAll(node.children);
  }

  return null;
}
