import 'dart:developer';

import 'package:flutter/rendering.dart';

/// An extension on [RenderObject] that provides some utility methods
extension RenderObjectExtension on RenderObject {
  /// Returns the frame Rect (position + size) of the element in the global
  /// coordinate system
  Rect? get globalFrame => _getFrameSafe(useGlobalCoordinates: true);

  /// Returns the frame Rect (position + size) of the element in the local
  /// coordinate system
  Rect? get localFrame => _getFrameSafe(useGlobalCoordinates: false);

  /// Returns the frame Rect (position + size) of the element
  ///
  /// This is "safe" to call on any RenderObject, even if it is not attached to
  /// the render tree, in which case it will return null.
  ///
  /// If [useGlobalCoordinates] is true, the frame is returned in the global coordinate
  /// system. Otherwise, it is returned in the local coordinate system.
  Rect? _getFrameSafe({required bool useGlobalCoordinates}) {
    final renderObject = this;

    // [RenderObject] is extended by 4 classes: [RenderBox], [RenderSliver],
    // [RenderView], and [RenderAbstractViewport].
    //
    // However [RenderAbstractViewport] does not implement [RenderObject.paintBounds]
    // therefore we can't implement a generic version of this method for it.
    if (renderObject is RenderView) {
      // according to the [RenderView] documentation, "it represents the root
      // element of the render tree", and has a single [RenderBox] as child
      // "which is required to fill the entire output surface." This is why it
      // makes sense to consider the frame of it's [RenderBox] child
      return renderObject.child?._getFrameSafe(
        useGlobalCoordinates: useGlobalCoordinates,
      );
    } else if (renderObject is RenderBox) {
      return _getFrameSafeForBox(
        renderObject,
        useGlobalCoordinates: useGlobalCoordinates,
      );
    } else if (renderObject is RenderSliver) {
      return _getFrameSafeForSliver(
        renderObject,
        useGlobalCoordinates: useGlobalCoordinates,
      );
    } else {
      log(
        'The RenderObject of type ${renderObject.runtimeType} is not supported. Therefore its size cannot be determined.',
      );
      return null;
    }
  }

  /// This is NOT safe to call if it was not verified that size of the render
  /// object does exist.
  ///
  /// Use [RenderObjectExtension.globalFrame] or [RenderObjectExtension.localFrame]
  /// instead.
  Rect _getFrameUnsafe({
    required bool useGlobalCoordinates,
  }) {
    if (!useGlobalCoordinates) {
      return paintBounds;
    }

    // If we want to use the global coordinates, we need to convert the local
    // coordinates to global coordinates by adding the offset applied to the
    // render object.
    final translation = getTransformTo(null).getTranslation();
    return paintBounds.shift(Offset(translation.x, translation.y));
  }
}

/// getFrameForBox
///
/// Returns the location Rect (position + size) of the given [RenderBox]
Rect? _getFrameSafeForBox(
  RenderBox renderBox, {
  required bool useGlobalCoordinates,
}) {
  if (!renderBox.hasSize) {
    return null;
  }

  return renderBox._getFrameUnsafe(
    useGlobalCoordinates: useGlobalCoordinates,
  );
}

/// getFrameForSliver
///
/// Returns the location Rect (position + size) of the given [RenderSliver]
Rect? _getFrameSafeForSliver(
  RenderSliver renderSliver, {
  required bool useGlobalCoordinates,
}) {
  if (renderSliver.geometry == null) {
    return null;
  }

  return renderSliver._getFrameUnsafe(
    useGlobalCoordinates: useGlobalCoordinates,
  );
}
