import 'package:flutter/widgets.dart';

/// An extension on [WidgetsBinding] that provides some utility methods
extension WidgetsBindingExtension on WidgetsBinding {
  /// This allows a value of type T or T? to be treated as a value of type T?.
  ///
  /// We use this so that APIs that have become non-nullable can still be used
  /// with `!` and `?` on the stable branch.
  static T? _ambiguate<T>(T? value) => value;

  /// A method to safely access [WidgetsBinding.instance]. This is needed because
  /// [WidgetsBinding.instance] is:
  ///   - nullable on Flutter < 3
  ///   - not nullable after Flutter >= 3
  ///
  /// We need to use the [_ambiguate] function because:
  ///   - Using a "!" on [instance] would trigger a warning for Flutter >= 3
  ///   - Not using a "!" would break the apps on Flutter < 3
  ///
  ///
  /// DON'T use [WidgetsBinding.instance] directly
  static WidgetsBinding? get safeInstance => _ambiguate(WidgetsBinding.instance);

}